FRAMERATE = 60
SCREEN_HEIGHT = 480
SCREEN_WIDTH = 1200

BACKGROUND_SPEED = 0.5
BACKGROUND_OPACITY = 125  # 0-255
BACKGROUND_WIDTH = 1700
BACKGROUND_HEIGHT = 600

TIME_LIMIT = 30

SHIP_SIZE = 75
SHIP_MAX_ANGLE = 45
